@@ -224,10 +224,11 @@ def cews(db: Dict[str, int],
          memo: Optional[dict] = {},
          wildcard: Optional[str] = '\uFFFF',
          max_wildcards: Optional[int] = 1,
-         min_samples_split: Optional[int] = 2,
+         min_samples_split: Optional[int] = None,
          min_samples_leaf: Optional[int] = 1,
          threshold: Optional[float] = 0.8,
-         approx_vocab_size: Optional[int] = None):
+         priority: Optional[str] = 'common',
+         vocab_size: Optional[int] = None):
     """Collectively Exhaustive Wildcard Shingling (CEWS)
 
     Parameters:
@@ -259,21 +260,34 @@ def cews(db: Dict[str, int],
         If an input string `s` contains more than the specified maximum
           number of wildcard characters, the recursion aborts.
 
-    min_samples_split: int (Default: 2)
+    min_samples_split: int (Default: None)
         If the combined frequency of all shingles covered by one wildcard
           shingle (count sum of the regex query results) is less than the
           specified minimum total frequency, then the recursion aborts.
+        If not set, then `min_samples_split = min_samples_leaf / threshold`
 
     min_samples_leaf: int (Default: 1)
-        Required number of counts in `db` to be added to the memo cached
+        Required number of counts in `db` to be added to the memo cached.
+        `min_samples_leaf` is estimated from `db` if
+        - it's a `float` between [0,1]
+        - `min_samples_leaf='auto'` and `vocab_size` is set
 
     threshold: float (Default: 0.80)
         Replace max. `1.0 - threshold` of the least frequent shingles with
           the wildcard shingle.
 
-    approx_vocab_size: int (Default: None)
-        If set, then `min_samples_leaf` is approximated by `approx_vocab_size`
-          and `db.values()`
+    priority: str (Default: 'common')
+        The order in which to process the shingles.
+        - 'common', start with the most frequent shingles. Use this setting
+            when specifying high `min_samples_leaf` and/or `vocab_size` to get
+            reduce the compute time
+        - 'rare', start with less frequent shingles. Use this setting when
+            when specifying low `min_samples_leaf=1` and not limiting the
+            vocab size.
+
+    vocab_size: int (Default: None)
+        Early stopping criteria. The main loop stops if `len(memo) >= vs`.
+          The actual memo cache size will be greater equal than `vocab_size`
 
     Return:
     -------
@@ -292,11 +306,10 @@ def cews(db: Dict[str, int],
     db_list = list(db.items())
     db_list = sorted(db_list, key=lambda item: -item[1])
 
-    # approximate `min_samples_leaf` with `approx_vocab_size`
-    if approx_vocab_size:
-        idx = min(len(db_list) - 1, approx_vocab_size * 2)
+    # approximate `min_samples_leaf` with `vocab_size`
+    if (min_samples_leaf == 'auto') and (vocab_size is not None):
+        idx = min(len(db_list) - 1, vocab_size * 2)
         min_samples_leaf = int(max(1, db_list[idx][1]))
-        min_samples_split = int(2 * min_samples_leaf)
 
     # if `min_samples_leaf` is float
     if isinstance(min_samples_leaf, float) and (0.0 < min_samples_leaf <= 1.0):
@@ -304,13 +317,25 @@ def cews(db: Dict[str, int],
         mask = (cnts / np.sum(cnts)) > min_samples_leaf
         if mask.any():
             min_samples_leaf = int(cnts[mask][-1])
-            min_samples_split = int(2 * min_samples_leaf)
         else:
-            min_samples_leaf, min_samples_split = 1, 2
+            min_samples_leaf = 1
+
+    # check threshold
+    assert threshold is not None
+    assert (0.0 < threshold <= 1.0)
+
+    # set `min_samples_split`
+    if min_samples_split is None:
+        min_samples_split = max(2, int(min_samples_leaf / threshold))
 
     # only shingles with at least `min_samples_leaf` counts
-    shingles = [s for s, c in db_list if c >= min_samples_leaf]
-    # shingles = [s for s, _ in db_list]
+    if priority == 'rare':
+        shingles = [s for s, _ in db_list]
+        shingles.reverse()
+    elif (priority == 'common') and (vocab_size is not None):
+        shingles = [s for s, c in db_list if c >= min_samples_leaf]
+    else:  # priority == 'common'
+        shingles = [s for s, _ in db_list]
 
     # loop over all db entries
     for s in shingles:
@@ -321,6 +346,11 @@ def cews(db: Dict[str, int],
             min_samples_split=min_samples_split,
             min_samples_leaf=min_samples_leaf,
             threshold=threshold)
+        # early stopping
+        if vocab_size is not None:
+            if len(memo) >= vocab_size:
+                break
+
     # done
     return memo
 
@@ -393,9 +423,13 @@ def shingles_to_patterns(memo: Dict[str, int],
         The regex.compile patterns based on the selected shingles in the
           memoization cache.
     """
+    if isinstance(memo, dict):
+        shingles = memo.keys()
+    elif isinstance(memo, (list, tuple)):
+        shingles = memo.copy()
     PATTERNS = []
-    for s in memo.keys():
-        reg = s.replace(wildcard, r"\w{1}")
+    for shingle in shingles:
+        reg = r"\w{1}".join([re.escape(s) for s in shingle.split(wildcard)])
         pat = re.compile(f"^{reg}$")
         PATTERNS.append(pat)
     PATTERNS.sort(key=sort_by_patterns)
